IMPLEMENTATION MODULE PakRdr;

FROM SYSTEM IMPORT BYTE, ADR;
IMPORT Lib2, Lib3, FIO, FIOStd, Types, SO, IO, Str;
IMPORT Config;
FROM Storage IMPORT ALLOCATE;

TYPE
  THdrDesc = RECORD
    id : ARRAY [0..3] OF CHAR;  (* 'PACK' *)
    ofs: Types.TLngCard;
    len: Types.TLngCard;
  END;

VAR
  pak: FIO.File;
  len: LONGCARD;
  hdr: THdrDesc;
  cnt: TCnt;
  toc: POINTER TO ARRAY TVldIdx OF TElmDesc;


PROCEDURE Fatal( t-: ARRAY OF CHAR );
VAR
  b: SO.TBuffer;
  s: ARRAY [0..127] OF CHAR;
BEGIN
  SO.Init( b, s );
  SO.WrStr( b, 'PakRdr.Fatal: ' );
  SO.WrStr( b, t );
  Lib3.FatalError( s );
  HALT;
END Fatal;

PROCEDURE NameValid( s-: ARRAY OF CHAR ): BOOLEAN;
VAR
  l, i: CARDINAL;
  c: CHAR;
BEGIN
  l := Str.Length( s );
  (* empty *)
  IF l = 0 THEN RETURN FALSE END;
  (* invalid char *)
  FOR i := 0 TO l - 1 DO
    c := s[i];
    IF ( c < ' ' ) OR ( c > '~' ) THEN RETURN FALSE END;
  END;
  (* !!! add more checks here !!! *)
  RETURN TRUE;
END NameValid;


PROCEDURE Cnt(): TCnt;
BEGIN
  RETURN cnt;
END Cnt;

PROCEDURE Elm( i: TVldIdx ): TElm;
BEGIN
  IF i > cnt THEN RETURN NIL END;
  RETURN ADR( toc^[i] );
END Elm;


PROCEDURE Pak();

  PROCEDURE Show();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'PakRdr.Pak: "' );
    SO.WrStr( b, Config.data.arc );
    SO.WrStr( b, '" len ' );
    b.c.p := '0';
    SO.WrLngHex( b, len, 8 );
    IO.WrStr( s ); IO.WrLn();
  END Show;

BEGIN
  IF ~ FIO.Exists( Config.data.arc ) THEN Fatal( 'Archive does not exist' ) END;
  pak := FIO.OpenRead( Config.data.arc );
  len := FIO.Size( pak );
  Show();
END Pak;

PROCEDURE Hdr();

  PROCEDURE Chr( c: CHAR ): CHAR;
  BEGIN
    IF ( c < ' ' ) OR ( '~' < c ) THEN c := '.' END;
    RETURN c;
  END Chr;

  PROCEDURE Show();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
    i: CARDINAL;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'PakRdr.Hdr: id "' );
    FOR i := 0 TO 3 DO
      SO.WrChar( b, Chr(hdr.id[i]) );
    END;
    SO.WrStr( b, '" ofs ' );
    b.c.p := '0';
    SO.WrLngHex( b, hdr.ofs, 8 );
    SO.WrStr( b, ' len ' );
    SO.WrLngHex( b, hdr.len, 8 );
    b.c.p := ' ';
    SO.WrStr( b, ' cnt ' );
    SO.WrLngCard( b, cnt, 1 );
    IO.WrStr( s ); IO.WrLn();
  END Show;

  PROCEDURE Check();
  BEGIN
    IF hdr.id # 'PACK'                  THEN Fatal( 'Invalid file ID' ) END;
    IF hdr.ofs < SIZE( hdr )            THEN Fatal( 'Invalid TOC offset (too small)' ) END;
    IF hdr.ofs > len                    THEN Fatal( 'Invalid TOC offset (too large)' ) END;
    IF hdr.len > len - hdr.ofs          THEN Fatal( 'Invalid TOC length (too large)' ) END;
    IF hdr.len MOD SIZE( TElmDesc ) # 0 THEN Fatal( 'Invalid TOC length (not multiple of 40H)' ) END;
  END Check;

BEGIN
  IF FIO.RdBin( pak, hdr, SIZE(hdr) ) # SIZE( hdr ) THEN Fatal( 'Archive too small' ) END;
  cnt := hdr.len DIV SIZE( TElmDesc );
  Show();
  Check();
END Hdr;

PROCEDURE Toc();
VAR
  l: CARDINAL;

  PROCEDURE Show();
  VAR
    b: SO.TBuffer;
    s: ARRAY [0..63] OF CHAR;
  BEGIN
    IF ~ Config.data.vrb THEN RETURN END;
    SO.Init( b, s );
    SO.WrStr( b, 'PakRdr.Toc: cnt ' );
    SO.WrCard( b, cnt, 1 );
    IO.WrStr( s ); IO.WrLn();
  END Show;

  PROCEDURE Check();
  VAR
    idx: CARDINAL;

    PROCEDURE Err( t-: ARRAY OF CHAR );
    VAR
      b: SO.TBuffer;
      s: ARRAY [0..127] OF CHAR;
    BEGIN
      SO.Init( b, s );
      SO.WrStr( b, 'Invalid TOC entry: idx ' );
      SO.WrLngCard( b, idx, 1 );
      SO.WrStr( b, ' ofs ' );
      b.c.p := '0';
      SO.WrLngHex( b, hdr.ofs + (idx-1) * SIZE(TElmDesc), 8 );
      SO.WrStr( b, ': ' );
      SO.WrStr( b, t );
      Fatal( s );
    END Err;

    PROCEDURE Elm( x-: TElmDesc );
    BEGIN
      IF ~ NameValid( x.str ) THEN Err( 'Bad file name' ) END;
      IF x.ofs > len THEN Fatal( 'Bad file offset' ) END;
      IF x.len > len - x.ofs THEN Fatal( 'Bad file length' ) END;
    END Elm;

  BEGIN
    FOR idx := 1 TO cnt DO Elm( toc^[idx] ) END;
  END Check;

BEGIN
  IF cnt > cntMax THEN Fatal( 'Too many TOC entries' ) END;
  l := cnt * SIZE( TElmDesc );
  ALLOCATE( toc, l );
  FIO.Seek( pak, hdr.ofs );
  l := FIO.RdBin( pak, toc^, l );
  IF l # hdr.len THEN Fatal( 'Cannot read TOC' ) END;
  Show();
  Check();
END Toc;

BEGIN
  pak := FIOStd.inv;
  len := 0;
  Lib2.Fill( hdr, 0 );
  cnt := 0;
  toc := NIL;

  Pak();
  Hdr();
  Toc();
END PakRdr.
